import json
from .exceptions import (ComponentHasNoEntity, ComponentAlreadyHaveEntity)

class Component():
    '''Base component class'''

    #__slots__ = ['_type', '_entity']

    default_component_type = None
    defaults = {}


    @property
    def type(self):
        return self._type

    @property
    def id(self):
        return self._id

    @property
    def entity(self):
        return self._entity

    @property
    def initialized(self):
        return self._entity


    def __init__(self, component_type=None, **properties):
        self._type = component_type or self.default_component_type or \
                                                            type(self).__name__
        self._id = 'no_id'
        self._entity = None

        if self.defaults == {}:
            for key, value in properties.items():
                #setattr(self, key, value)
                self.defaults[key] = value

        for key, value in self.defaults.items():
            setattr(self, key, properties.pop(key, value))

    def __repr__(self):
        # <Component:type; Enity:entity>
        return "<Component:{}; Entity:{}>".format(self._type,
                                            self._entity if self._entity is None
                                                        else self._entity.type)

    def __str__(self):
        # JSON
        keys = self.defaults.keys()
        data = {}

        for key in keys:
            data[key] = getattr(self, key)

        json_string = json.dumps(data, indent=4)
        return json_string


    def create(self, entity):
        if self.initialized:
            raise ComponentAlreadyHaveEntity(self)

        self._entity = entity

    def destroy(self):
        if not self.initialized:
            raise ComponentHasNoEntity(self)

        entity = self._entity
        if entity.initialized:
            entity._component_manager.destroy_component(self)
        else:
            self._entity = None



from copy import deepcopy
from .exceptions import (EntityNotCreated, EntityAlreadyExists,
                         ComponentAlreadyHaveEntity)


class Entity():
    '''Base entity class'''

    __slots__ = ['_type', '_id', '_name', '_components',
                 '_entity_manager', '_component_manager']

    default_entity_type = None
    default_components = {}


    @property
    def type(self):
        return self._type

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def components(self):
        return self._components

    @property
    def initialized(self):
        return self._entity_manager and self._component_manager


    def __init__(self, name, entity_type=None):
        self._type = entity_type or self.default_entity_type or \
                                                            type(self).__name__

        self._id = 'no_id'
        self._name = name
        self._components = deepcopy(self.default_components)

        for component in self._components.values():
            component.create(self)

        self._entity_manager = None
        self._component_manager = None

    def __repr__(self):
        return "<Entity:{}; {}:{}>".format(self._type, self._name, self._id)

    def __str__(self):
        ent = "<Entity:{}; {}:{}>".format(self._type, self._name, self._id)
        components = self._components
        # TO DO: return repr of component
        # repr(component)
        return "{}\n{}".format(ent, components)

    def __getitem__(self, key):
        return self._components[key]

    def __setitem__(self, key, value):
        if value.initialized:
            raise ComponentAlreadyHaveEntity(value)

        old_component = self._components.pop(key, None)
        if old_component is not None:
            old_component.destroy()

        self._components[key] = value
        value.create(self)

        if self.initialized:
            self._component_manager.create_component(value)
            
    def __delitem__(self, key):
        component = self._components.pop(key)
        component.destroy()

    def __getattr__(self, key):
        if key in super().__getattribute__('__slots__'):
            return super().__getattr__(key)
        else:
            return self._components[key]

    def __setattr__(self, key, value):
        if key in super().__getattribute__('__slots__'):
            super().__setattr__(key, value)
        else:
            if value.initialized:
                raise ComponentAlreadyHaveEntity(value)

            old_component = self._components.pop(key, None)
            if old_component is not None:
                old_component.destroy()

            self._components[key] = value
            value.create(self)

            if self.initialized:
                self._component_manager.create_component(value)
                
    def __delattr__(self, key):
        component = self._components.pop(key)
        component.destroy()


    def create(self, entity_manager, component_manager):
        if self.initialized:
            raise EntityAlreadyExists(self)

        self._entity_manager = entity_manager
        self._component_manager = component_manager

        self._entity_manager.create_entity(self)
        for component in self._components.values():
            self._component_manager.create_component(component)

    def destroy(self):
        if not self.initialized:
            raise EntityNotCreated(self)

        for component in self._components.values():
            self._component_manager.destroy_component(component)
        self._entity_manager.destroy_entity(self)

        self._entity_manager = None
        self._component_manager = None



from .event import Event
from .entity import Entity
from .component import Component
from .system import System

from .exceptions import (EntityNotCreated, EntityAlreadyExists,
                         EntityAlredyHaveComponent,
                         ComponentHasNoEntity, ComponentAlreadyHaveEntity,
                         SystemNotCreated, SystemAlreadyExists,
                         FunctionNotSubscribed, FunctionAlreadySubscribed)



from .models import Event
from .models import Entity
from .models import Component
from .models import System

from .managers import EventManager
from .managers import EntityManager
from .managers import ComponentManager
from .managers import SystemManager

from .engine import Engine

from .models import (EntityNotCreated, EntityAlreadyExists,
                     EntityAlredyHaveComponent,
                     ComponentHasNoEntity, ComponentAlreadyHaveEntity,
                     SystemNotCreated, SystemAlreadyExists,
                     FunctionNotSubscribed, FunctionAlreadySubscribed)



from ..models import (SystemNotCreated, SystemAlreadyExists)


class SystemManager():
    '''System manager'''

    @property
    def system_types(self):
        return self._system_types


    def __init__(self):
        self._system_types = {}

    def create_system(self, system):
        if system.type in self._system_types.keys():
            raise SystemAlreadyExists(system)

        self._system_types[system.type] = system

    def destroy_system(self, system):
        if system.type not in self._system_types.keys():
            raise SystemNotCreated(system)

        del self._system_types[system.type]



class EntityAlreadyExists(Exception):

    def __init__(self, entity):
        self.entity = entity

    def __str__(self):
        return "Entity {} already exist".format(self.entity)

class EntityNotCreated(Exception):

    def __init__(self, entity):
        self.entity = entity

    def __str__(self):
        return "Entity {} not created".format(self.entity)

class EntityAlredyHaveComponent(Exception):

    def __init__(self, entity, component):
        self.entity = entity
        self.component = component

    def __str__(self):
        return "Entity {} already has a component {}".format(self.entity,
                                                             self.component)

class ComponentAlreadyHaveEntity(Exception):

    def __init__(self, component):
        self.component = component

    def __str__(self):
        return "Component {} already have entity {}".format(self.component,
                                                        self.component.entity)

class ComponentHasNoEntity(Exception):

    def __init__(self, component):
        self.component = component

    def __str__(self):
        return "Component {} has no entity".format(self.component)

class SystemAlreadyExists(Exception):

    def __init__(self, system):
        self.system = system

    def __str__(self):
        return "System {} already exists".format(self.system)

class SystemNotCreated(Exception):

    def __init__(self, system):
        self.system = system

    def __str__(self):
        return "System {} not created".format(self.system)

class FunctionAlreadySubscribed(Exception):

    def __init__(self, fn, event_type):
        self.fn = fn
        self.event_type = event_type

    def __str__(self):
        return "Function {} already subscribed to the event {}".format(self.fn,
                                                                self.event_type)

class FunctionNotSubscribed(Exception):

    def __init__(self, fn, event_type):
        self.fn = fn
        self.event_type = event_type

    def __str__(self):
        return "Function {} not subscribed to the event {}".format(self.fn,
                                                                self.event_type)



from .event import Event
from .exceptions import (SystemNotCreated, SystemAlreadyExists)


class System():
    '''Base system class'''

    __slots__ = ['_type', '_entity_manager', '_component_manager',
                 '_system_manager', '_event_manager']

    default_system_type = None


    @property
    def type(self):
        return self._type

    @property
    def initialized(self):
        return self._entity_manager and self._component_manager and \
               self._system_manager and self._event_manager

    @property
    def entity_manager(self):
        return self._entity_manager

    @property
    def component_manager(self):
        return self._component_manager

    @property
    def system_manager(self):
        return self._system_manager

    @property
    def event_manager(self):
        return self._event_manager


    def __init__(self, system_type=None):
        self._type = system_type or self.default_system_type or type(self).__name__

        self._entity_manager = None
        self._component_manager = None
        self._system_manager = None
        self._event_manager = None

        self._subscribed_functions = []


    def create_entity(self, entity):
        if not self.initialized:
            raise SystemNotCreated(self)

        entity.create(self._entity_manager, self._component_manager)

    def generate_event(self, event_type, data=None):
        if not self.initialized:
            raise SystemNotCreated(self)

        event = Event(event_type, data)
        self._event_manager.add_event(event)

    def send_event(self, event):
        if not self.initialized:
            raise SystemNotCreated(self)
    
        self._event_manager.add_event(event)

    def subscribe(self, fn, event_type):
        # TO DO: decorator
        if not self.initialized:
            raise SystemNotCreated(self)

        self._subscribed_functions.append((fn, event_type))
        self._event_manager.subscribe(fn, event_type)

    def unsubscribe(self, fn, event_type):
        if not self.initialized:
            raise SystemNotCreated(self)

        self._subscribed_functions.remove((fn, event_type))
        self._event_manager.unsubscribe(fn, event_type)

    def create(self, entity_manager, component_manager,
                                                system_manager, event_manager):
        if self.initialized:
            raise SystemAlreadyExists(self)

        self._entity_manager = entity_manager
        self._component_manager = component_manager
        self._system_manager = system_manager
        self._event_manager = event_manager

        self._system_manager.create_system(self)

        self.init()

    def destroy(self):
        self.final()

        if not self.initialized:
            raise SystemNotCreated(self)

        for subscriber in self._subscribed_functions:
            self._event_manager.unsubscribe(subscriber[0], subscriber[1])
        self._system_manager.destroy_system(self)

        self._entity_manager = None
        self._component_manager = None
        self._system_manager = None
        self._event_manager = None


    def init(self):
        pass

    def final(self):
        pass



from ..models import (EntityNotCreated, EntityAlreadyExists)


class EntityManager():
    '''Entity manager'''

    @property
    def entities(self):
        return self._entities

    @property
    def entity_types(self):
        return self._entity_types


    def __init__(self):
        self._entity_count = 0
        self._entities = []
        self._entity_types = {}

    def create_entity(self, entity):
        if entity in self._entities:
            raise EntityAlreadyExists(entity)

        entity._id = self._entity_count
        self._entity_count+=1

        self._entities.append(entity)

        if self._entity_types.get(entity.type) is None:
            self._entity_types[entity.type] = {}
        self._entity_types[entity.type][entity.id] = entity

    def destroy_entity(self, entity):
        if entity not in self._entities:
            raise EntityNotCreated(entity)

        self._entities[entity.id] = None

        del self._entity_types[entity.type][entity.id]
        if self._entity_types[entity.type] == {}:
            del self._entity_types[entity.type]



from .models import Event
from .managers import (EntityManager, ComponentManager,
                       SystemManager, EventManager)

class Engine():

    @property
    def entity_manager(self):
        return self._entity_manager

    @property
    def component_manager(self):
        return self._component_manager

    @property
    def system_manager(self):
        return self._system_manager

    @property
    def event_manager(self):
        return self._event_manager

    def __init__(self):
        self._entity_manager = EntityManager()
        self._component_manager = ComponentManager()
        self._system_manager = SystemManager()
        self._event_manager = EventManager()

    def create_system(self, system):
        system.create(self._entity_manager, self._component_manager,
                      self._system_manager, self._event_manager)

    def generate_input(self, event_type='InputEvent', data=None):
        event = Event(event_type, data)

        self._event_manager.add_event(event)

    def send_input(self, event):
        self._event_manager.add_event(event)

    def get_output(self):
        events = self._event_manager.events
        output = []

        for event in events:
            if event.type == 'OutputEvent':
                output.append(event)

        return output



from ..models import (ComponentHasNoEntity, ComponentAlreadyHaveEntity,
                      EntityAlredyHaveComponent)


class ComponentManager():
    '''Component manager'''

    @property
    def components(self):
        return self._components

    @property
    def component_types(self):
        return self._component_types


    def __init__(self):
        self._component_count = 0
        self._components = []
        self._component_types = {}

    def create_component(self, component):
        if component in self._components:
            # TO DO: fix this error
            raise ComponentAlreadyHaveEntity(component)

        component._id = self._component_count
        self._component_count+=1

        self._components.append(component)

        if self._component_types.get(component.type) is None:
            self._component_types[component.type] = {}

        if self._component_types[component.type].get(component.entity.id) \
                                                                    is not None:
            EntityAlredyHaveComponent(component.entity, component)

        self._component_types[component.type][component.entity.id] = component

    def destroy_component(self, component):
        if component not in self._components:
            raise ComponentHasNoEntity(component)

        self._components[component.id] = None

        del self._component_types[component.type][component.entity.id]
        if self._component_types[component.type] == {}:
            del self._component_types[component.type]
            
        component._entity = None



class Event():
    '''Event class'''


    __slots__ = ['_type', '_data']

    default_event_type = None
    default_event_data = {}


    @property
    def type(self):
        return self._type

    @property
    def data(self):
        return self._data


    def __init__(self, type=None, data=None):
        self._type = type or self.default_event_type or type(self).__name__
        self._data = data or self.default_event_data

    def __repr__(self):
        # <Event:event_type>
        return "<Event:{}".format(self.type)

    def __str__(self):
        # <Event:event_type>
        # {data}
        return "<Event:{}>\n{}".format(self.type, self.data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        self._data.pop(key)

    def __getattr__(self, key):
        if key in super().__getattribute__('__slots__'):
            return super().__getattr__(key)
        else:
            return self._data[key]

    def __setattr__(self, key, value):
        if key in super().__getattribute__('__slots__'):
            super().__setattr__(key, value)
        else:
            self._data[key] = value

    def __delattr__(self, key):
        self._data.pop(key)



from setuptools import setup, find_packages

setup(name='edecs',
      version='0.1.11',
      description='Event-driven Entity-Component-System engine',
      packages=find_packages(),
      author_email='dovakin121212@gmail.com',
      zip_safe=False)



from ..models import (FunctionNotSubscribed, FunctionAlreadySubscribed)


class EventManager():
    '''Event manager'''

    @property
    def subscribers(self):
        return self._subscribers

    @property
    def events(self):
        return self._events

    @property
    def empty(self):
        return self._events == []


    def __init__(self):
        self._subscribers = {}
        self._events = []

    def subscribe(self, fn, event_type):
        if self._subscribers.get(event_type) is None:
            self._subscribers[event_type] = []

        if fn in self._subscribers[event_type]:
            raise FunctionAlreadySubscribed(fn, event_type)

        self._subscribers[event_type].append(fn)

    def unsubscribe(self, fn, event_type):
        if self._subscribers.get(event_type) is None or \
                                        fn not in self._subscribers[event_type]:
            raise FunctionNotSubscribed(fn, event_type)

        self._subscribers[event_type].remove(fn)

        if self._subscribers[event_type] == {}:
            del self._subscribers[event_type]

    def add_event(self, event):
        self._events.append(event)

    def update_events(self):
        if self._events != []:
            event = self._events.pop(0)
            subscribers = self._subscribers.get(event.type)
            if subscribers is not None:
                for subscriber in subscribers:
                    subscriber(event)
